ifndef NBODY_ACCELERATION_HH
#define NBODY_ACCELERATION_HH

#include <vector>
#include <cassert>
#include <cmath>
#include <limits>

#include "common/timefunction.hh"
#include "nbody_datastructure.hh"
#include "nbody_vectoroperations.hh"
#include "nbody_physics.hh"


/**
 * Right hand side function f of the n-body initial value problem of second
 * order
 *   r'' = f(r) on [t_0,T], r(t_0) = r_0, r'(t_0) = v_0,
 * with f: (R^2)^n --> (R^2)^n, t_0,T \in R_0^+, r_0,v_0 \in (R^2)^n, n \in N.
 * This function computes a vector a := f(r) containing the acceleration of
 * the bodies.
 * TODO (in future release):
 *   Allocate temporary variables (and constants) only once by turning
 *   them into class members; make heavy use of /= and -= operations to
 *   save allocation of temporary Vector2D instances.
 */
class NBody_Acceleration
  : public TimeFunction<Body,
                        Vector2D,
                        double>
{
public:
  /**
   * Creates an instance of the right hand side function.
   * @param[in] data Vector containing the mass of the bodies.
   */
  NBody_Acceleration (const DataVariant1& data
                    , const double theta = 0.5
                    , const double epsilon2 = 1e-3)
  : data_(data), theta_(theta), epsilon2_(epsilon2)
  { 
  }


  void setupTree(const DataVariant1& data)
  {
  }

  /**
   * Compute acceleration vector with n^2 algorithm.
   * @param[in]  t Point in temporal domain;
   *               not used here as we solve an autonomous problem.
   * @param[in]  r Vector containing the position of the bodies.
   * @param[out] a Vector containing the acceleration of the bodies.
   */
  virtual void evaluate (const typename NBody_Acceleration::TimeType& /*t*/,
                         const typename NBody_Acceleration::DomainType& b,
                         typename NBody_Acceleration::RangeType& a) const
  {
      /**
       * Using Newton's 3rd law, we compute the force vector F_ij only
       * once for each pair (i,j) and let it directly act on both bodies.
       * As there is no self-interaction we meanwhile exclude the case
       * (i == j).
       */
   }

private:
  
  const DataVariant1& data_;
  QuadTreeNode root;
  const double theta_;
  const double epsilon2_;
};


#endif // NBODY_ACCELERATION_HH
