#ifndef NBODY_VECTOROPERATIONS_HH
#define NBODY_VECTOROPERATIONS_HH

#include <cmath>

#include "nbody_datastructure.hh"


// ********************** function declarations *********************
// overload the C++ arithmetic operators to support vector operations
// ******************************************************************


/**
 * Computes the sum of two vectors in R^2.
 * @param[in,out] lhs Vector in R^2;
 *                    contains first summand and the result.
 * @param[in]     rhs Vector in R^2.
 * @return Reference to lhs (after summation).
 */
Vector2D& operator+= (Vector2D& lhs, const Vector2D& rhs);


/**
 * Computes the difference of two vectors in R^2.
 * @param[in,out] lhs Vector in R^2;
 *                    contains minuend and the result.
 * @param[in]     rhs Vector in R^2.
 * @return Reference to lhs (after subtraction).
 */
Vector2D& operator-= (Vector2D& lhs, const Vector2D& rhs);


/**
 * Computes scalar multiplication in R^2,
 * i.e. the multiplication of a vector in R^2 by a scalar in R.
 * @param[in,out] lhs Vector in R^2;
 *                contains input vector and the result.
 * @param[in]     rhs Scalar in R.
 * @return Reference to lhs (after scalar multiplication).
 */
Vector2D& operator*= (Vector2D& lhs, const double& rhs);


/**
 * Computes the sum of two vectors in R^2.
 * @param[in] lhs Vector in R^2.
 * @param[in] rhs Vector in R^2.
 * @return Vector containing the result lhs+rhs.
 */
const Vector2D operator+ (const Vector2D& lhs, const Vector2D& rhs);


/**
 * Computes the difference of two vectors in R^2.
 * @param[in] lhs Vector in R^2.
 * @param[in] rhs Vector in R^2.
 * @return Vector containing the result lhs-rhs.
 */
const Vector2D operator- (const Vector2D& lhs, const Vector2D& rhs);


/**
 * Computes scalar multiplication in R^2,
 * i.e. the multiplication of a vector in R^2 by a scalar in R.
 * @param[in] lhs Scalar in R.
 * @param[in] rhs Vector in R^2.
 * @return Vector containing the result lhs*rhs.
 */
const Vector2D operator* (const double& lhs, const Vector2D& rhs);


/**
 * Computes scalar multiplication in R^2,
 * i.e. the multiplication of a vector in R^2 by a scalar in R.
 * @param[in] lhs Vector in R^2.
 * @param[in] rhs Scalar in R.
 * @return Vector containing the result rhs*lhs.
 */
const Vector2D operator* (const Vector2D& lhs, const double& rhs);


/**
 * Multiplies a vector in R^2 by the multiplicative inverse of a
 * scalar in R.
 * @param[in] lhs Vector in R^2.
 * @param[in] rhs Scalar in R.
 * @return Vector containing the result (1/rhs)*lhs.
 */
const Vector2D operator/ (const Vector2D& lhs, const double& rhs);

/**
 * Compares two vectors in R^2 with respect to machine precision.
 * @param[in] lhs Vector in R^2.
 * @param[in] rhs Vector in R^2.
 * @return true if  rel_fehler(lhs - rhs) < 1e-16.
 */
const bool operator== (const Vector2D& lhs, const double& rhs);


// ********************** function definitions **********************
// we also put the functions' definitions (their implementation) into
// this header file as we want the compiler to inline these functions
// ******************************************************************


inline Vector2D& operator+= (Vector2D& lhs, const Vector2D& rhs)
{
  for (unsigned int i = 0; i < lhs.size(); ++i)
    lhs[i] += rhs[i];

  return lhs;
}


inline Vector2D& operator-= (Vector2D& lhs, const Vector2D& rhs)
{
  for (unsigned int i = 0; i < lhs.size(); ++i)
    lhs[i] -= rhs[i];

  return lhs;
}


inline Vector2D& operator*= (Vector2D& lhs, const double& rhs)
{
  for (unsigned int i = 0; i < lhs.size(); ++i)
    lhs[i] *= rhs;

  return lhs;
}


inline const Vector2D operator+ (const Vector2D& lhs, const Vector2D& rhs)
{
  // copy left operand
  Vector2D result(lhs);

  // implement via operator+= for Vector2D
  result += rhs;

  return result;
}


inline const Vector2D operator- (const Vector2D& lhs, const Vector2D& rhs)
{
  // copy left operand
  Vector2D result(lhs);

  // implement via operator-= for Vector2D
  result -= rhs;

  return result;
}


inline const Vector2D operator* (const double& lhs, const Vector2D& rhs)
{
  // copy right operand
  Vector2D result(rhs);

  // implement via operator*= for Vector2D
  result *= lhs;

  return result;
}


inline const Vector2D operator* (const Vector2D& lhs, const double& rhs)
{
  // implement via "scalar times vector" operator
  return rhs * lhs;
}


inline const Vector2D operator/ (const Vector2D& lhs, const double& rhs)
{
  // implement via "scalar times vector" operator
  return (1.0 / rhs) * lhs;
}


inline const bool operator== (const Vector2D& lhs, const Vector2D& rhs)
{
  // implement via "scalar times vector" operator
  return (std::fabs(lhs[0]-rhs[0])/std::fabs(lhs[0]+rhs[0]))<1e-16 && (std::fabs(lhs[1]-rhs[1])/std::fabs(lhs[1]+rhs[1]))<1e-16;
}


#endif // NBODY_VECTOROPERATIONS_HH
